@@ -33,6 +33,7 @@ static struct mtu_def mtu[] = {
   {  1400, "(Google/AOL)" }, 	/* To be investigated */
   {  1415, "sometimes wireless" },
   {  1420, "GPRS or FreeS/WAN" },
+  {  1423, "sometimes cable" },
   {  1440, "sometimes DSL (2)" },
   {  1442, "IPIP tunnel" },
   {  1450, "vtun" },
@@ -41,8 +42,9 @@ static struct mtu_def mtu[] = {
   {  1456, "ISDN ppp" },
   {  1462, "sometimes DSL (5)" },
   {  1476, "IPSec/GRE" },
-  {  1480, "IPv6" },
+  {  1480, "IPv6/IPIP" },
   {  1492, "pppoe (DSL)" },
+  {  1496, "vLAN" },
   {  1500, "ethernet/modem" },
   {  1656, "Ericsson HIS" },
   {  2024, "wireless/IrDA" },
@@ -51,6 +53,7 @@ static struct mtu_def mtu[] = {
   {  3924, "loopback" },
   {  4056, "token ring (1)" },
   {  4096, "Sangoma X.25 WAN" },
+  {  4352, "FDDI" },
   {  4500, "token ring (2)" },
   {  9180, "FORE ATM" },
   { 16436, "sometimes loopback" },
